local _G = _G
local date = _G.date
local MAJOR_VERSION = 'GraveLevelBar'

local LibStub = _G.LibStub
local AceAddon = LibStub('AceAddon-3.0')
local addon = AceAddon:GetAddon(MAJOR_VERSION)

local min = min

addon.xpBarTemplate = {}
local xpBarTemplate = addon.xpBarTemplate

xpBarTemplate.defaults = {
    startXP = 0,
    targetXP = nil,
}

xpBarTemplate.optionsTable = {
    name = nil,
    type = 'group',
    args = {
        startXP = {
            name = "Start XP",
            desc = "Minimum XP value to display.",
            type = 'range',
            min = 0,
            max = nil,
            step = 1,
            bigStep = nil,
            get = 'getStartXP',
            set = 'setStartXP',
        },
        startLevel = {
            name = "Start Level",
            desc = "Minimum level to display.",
            type = 'range',
            min = 1,
            max = nil,
            bigStep = 1,
            get = 'getStartLevel',
            set = 'setStartLevel',
        },
        targetXP = {
            name = "Target XP",
            desc = "Maximum XP value to disply.",
            type = 'range',
            min = nil,
            max = nil,
            step = 1,
            bigStep = nil,
            get = 'getTargetXP',
            set = 'setTargetXP',
        },
        targetLevel = {
            name = "Target Level",
            desc = "Maximum level to disply.",
            type = 'range',
            min = nil,
            max = nil,
            bigStep = 1,
            get = 'getTargetLevel',
            set = 'setTargetLevel',
        },
    },
}

-- start xp
function xpBarTemplate:getStartXP(info)
    local startXP = self:getOptions().startXP
    local startXPMax = self:getStartXPMax()
    if startXP > self:getStartXPMax() then
        return startXPMax
    end
    return self:getOptions().startXP
end

function xpBarTemplate:getStartXPMin()
    error("Implement me")
end

function xpBarTemplate:getStartXPMax()
    return self:getCurrentXP()
end

function xpBarTemplate:setStartXP(info,value)
    local startXPMin = self:getStartXPMin()
    local startXPMax = self:getStartXPMax()
    if value < startXPMin then
        value = startXPMin
    end
    if value > startXPMax then
        value = startXPMax
    end
    self:getOptions().startXP = value
    self:Invalidate()
end

-- start level
function xpBarTemplate:getStartLevel(info)
    local startXP = self:getStartXP()
    local level = self:calcLevelForXP(startXP)
    return level
end

function xpBarTemplate:getStartLevelMin()
    local startXPMin = self:getStartXPMin()
    local levelMin = self:calcLevelForXP(startXPMin)
    return levelMin
end

function xpBarTemplate:getStartLevelMax()
    local startXPMax = self:getStartXPMax()
    local levelMax = self:calcLevelForXP(startXPMax)
    return levelMax
end

function xpBarTemplate:setStartLevel(info,value)
    local prevLevel = floor(value)
    local prevLevelXP = self:calcXpForLevel(prevLevel)
    local nextLevel = ceil(value)
    if prevLevel < nextLevel then
        local nextLevelXP = self:calcXpForLevel(nextLevel)
        local startXP =  prevLevelXP + (value-prevLevel)*(nextLevelXP-prevLevelXP)
        self:setStartXP(info,startXP)
    else
        self:setStartXP(info,prevLevelXP)
    end
end

-- target xp
function xpBarTemplate:getTargetXP(info)
    return self:getOptions().targetXP or self:getTargetXPMax()
end

function xpBarTemplate:getTargetXPMin()
    return self:getCurrentXP()
end

function xpBarTemplate:getTargetXPMax()
    local maxPlayerLevel = self:getTargetLevelMax()
    return self:calcXpForLevel(maxPlayerLevel)
end

function xpBarTemplate:setTargetXP(info,value)
    self:getOptions().targetXP = value
    self:Invalidate()
end

-- target level
function xpBarTemplate:getTargetLevel(info)
    local targetXP = self:getTargetXP()
    local level = self:calcLevelForXP(targetXP)
    return level
end

function xpBarTemplate:getTargetLevelMin()
    local targetXPMin = self:getTargetXPMin()
    local levelMin = self:calcLevelForXP(targetXPMin)
    return levelMin
end

function xpBarTemplate:getTargetLevelMax()
    local targetXPMax = self:getTargetXPMax()
    local levelMax = self:calcLevelForXP(targetXPMax)
    return levelMax
end

function xpBarTemplate:OnInitialize()
    self.superOnInitialize(self)
    self:updateOptionsTable()
end

function xpBarTemplate:OnEnable()
    self.superOnEnable(self)
    self:updateOptionsTable()
end

function xpBarTemplate:setTargetLevel(info,value)
    assert(self and self.optionsTable and self.optionsTable.name, "where is my optionsTable?")
    assert(value and type(value) == 'number',"suplied argument value is not a number: name=" .. tostring(self.optionsTable.name) .. ", value=" .. tostring(value))
    local prevLevel = floor(value)
    local prevLevelXP = self:calcXpForLevel(prevLevel)
    local nextLevel = ceil(value)
    assert(prevLevel,"prev level is nil. name=" .. tostring(self.optionsTable.name) .. " value=" .. tostring(value))
    assert(nextLevel,"next level is nil. name=" .. tostring(self.optionsTable.name) .. " value=" .. tostring(value))
    if prevLevel < nextLevel then
        local nextLevelXP = self:calcXpForLevel(nextLevel)
        local targetXP =  prevLevelXP + (value-prevLevel)*(nextLevelXP-prevLevelXP)
        self:setTargetXP(info,targetXP)
    else
        self:setTargetXP(info,prevLevelXP)
    end
end

-- end of setters and getters


function xpBarTemplate:updateOptionsTable()
    --update options table
    local startXPMin = self:getStartXPMin()
    local startXPMax = self:getStartXPMax()
    self.optionsTable.args.startXP.min = startXPMin
    self.optionsTable.args.startXP.max = startXPMax
    self.optionsTable.args.startLevel.max = self:getStartLevelMax()
    self.optionsTable.args.targetXP.max = self:getTargetXPMax()
    self.optionsTable.args.targetLevel.max = self:getTargetLevelMax()
    self.optionsTable.args.targetXP.min = self.optionsTable.args.startXP.max
    self.optionsTable.args.targetLevel.min = self.optionsTable.args.startLevel.max
    assert(startXPMin <= startXPMax, "startXPMin=" .. tostring(startXPMin) .. ", startXPMax=" .. tostring(startXPMax))
    assert(self.optionsTable.args.startLevel.min  <= self.optionsTable.args.startLevel.max)
    assert(self.optionsTable.args.targetXP.min    <= self.optionsTable.args.targetXP.max)
    assert(self.optionsTable.args.targetLevel.min <= self.optionsTable.args.targetLevel.max)
end

function xpBarTemplate:getCurrentXP()
    error("Not implemented")
end

function xpBarTemplate:getCurrentLevel()
    return self:calcLevelForXP(self:getCurrentXP())
end

--
local function drawElement(cache, frame, position, level, baseFontSize)
    -- texture
    local levelTexture = cache.texture
    if not levelTexture then
        --levelCache[level].texture = frame:CreateTexture(frame:GetName()..'Level'..level,'OVERLAY')
        cache.texture = frame:CreateTexture(nil,'OVERLAY')
        levelTexture = cache.texture
        levelTexture:SetWidth(1)
        levelTexture:SetTexture(1,1,1,1)
    end
    levelTexture:ClearAllPoints()
    levelTexture:SetPoint('TOP')
    levelTexture:SetPoint('BOTTOM')
    levelTexture:SetPoint('LEFT',frame,'LEFT', position,0)
    levelTexture:Show()

    -- string
    local levelString = cache.string
    if not levelString then
        cache.string = frame:CreateFontString(nil,'OVERLAY')
        levelString = cache.string
        -- calculate font size
        local digits = string.len(tostring(level)) - 1
        --addon:Debug(level,digits)
        local size = max(baseFontSize - digits * 2,4)

        local isValid = levelString:SetFont('Fonts\\FRIZQT__.TTF',size)
        assert(isValid)
        levelString:SetJustifyV('MIDDLE')
        levelString:SetText(level)
    end
    levelString:ClearAllPoints()
    if position > 20 then
        levelString:SetPoint('RIGHT',cache.texture,'LEFT')
        levelString:SetJustifyH('RIGHT')
    else
        levelString:SetPoint('LEFT',cache.texture,'RIGHT')
        levelString:SetJustifyH('LEFT')
    end
    levelString:Show()
end

local function hideElement(cache)
    cache.texture:Hide()
    cache.string:Hide()
end

local function hideAll(t)
    if t.visibleLevels then
        for cache in pairs(t.visibleLevels) do
            hideElement(cache)
        end
    end
end

-- start is inclusive, end is exclusive
function xpBarTemplate:DrawNumbersSub(startPos, startLevel, endPos, endLevel)
    local barCache = self.cache
    local frame = barCache.frame
    local width = frame:GetRight() - frame:GetLeft()

    -- some cache
    barCache.levels[startLevel] = barCache.levels[startLevel] or {}
    local levelCache = barCache.levels[startLevel]

    if startPos > width or endPos <= 0 then
        hideAll(levelCache)
        --addon:Debug("fail", startLevel, startPos,endLevel, endPos)
        return
    --else
        --addon:Debug("pass", startLevel, startPos,endLevel, endPos)
    end

    local fontSize = self:getOptions().size
    if startLevel < endLevel then
        local currentWidth = endPos - startPos
        local displayLevels  = floor(log10(currentWidth/50))
        local steps =10^displayLevels
        local levelStepSize = 1 / steps
        --local stepSize = currentWidth/steps
        local levels = endLevel - startLevel

        local drawn = {}
        levelCache.visibleLevels = levelCache.visibleLevels or {}
        local visible = levelCache.visibleLevels

        local l = startLevel
        while l < endLevel do
            local pos = startPos + (l-startLevel)/levels * currentWidth

            if 0 <= pos and pos <= width then
                levelCache[l] = levelCache[l] or {}
                drawElement(levelCache[l], frame, pos, l, fontSize)
                visible[levelCache[l]] = true
                drawn[levelCache[l]] = true
                --addon:Debug("Drawn:",l,drawn[levelCache[l]])
            else
                --addon:Debug("Not Drawn:",l)
            end

            -- post increment of loop
            l = l + levelStepSize
        end

        -- hide old
        for cache,_ in pairs(visible) do
            if not drawn[cache] then
                hideElement(cache)
                visible[cache] = nil
                --addon:Debug("Hide:",cache)
            else
                --addon:Debug("Don't hide:",cache)
            end
        end
    else
        -- single element
        if 0 <= endPos and endPos < width then
            levelCache[endLevel] = levelCache[endLevel] or {}
            drawElement(levelCache[endLevel], frame, endPos, endLevel, fontSize)
        elseif levelCache[endLevel] then
            hideElement(levelCache[endLevel])
        end
    end
end

function xpBarTemplate:DrawNumbers()
    local startLevel = floor(self:getStartLevelMin())
    local targetLevel = ceil(self:getTargetLevelMax())
    local options = self:getOptions()
    local size = options.size
    local barCache = self.cache
    barCache.levels = barCache.levels or {}
    local levelCache = barCache.levels
    assert(type(startLevel) == 'number', "startLevel is a " .. type(startLevel))

    local frame = barCache.frame
    local width = frame:GetRight() - frame:GetLeft()

    local startGlobalXP = self:getStartXP()
    local targetGlobalXP = self:getTargetXP()
    for level = startLevel,targetLevel do
        local startXpForLevel = self:calcXpForLevel(level)
        local endXpForLevel = level < targetLevel and self:calcXpForLevel(level+1) or startXpForLevel
        local startRatio = (startXpForLevel-startGlobalXP)/(targetGlobalXP-startGlobalXP)
        local endRatio =   (endXpForLevel  -startGlobalXP)/(targetGlobalXP-startGlobalXP)
        local curPos = startRatio * width
        local nextPos = endRatio * width
        self:DrawNumbersSub(curPos, level, nextPos, level+1)
    end
end

function addon:createXPModule(name,...)
    assert(addon.barTemplate.optionsTable.type)
    local template = addon:merge(addon.barTemplate, xpBarTemplate)
    assert(template.optionsTable.type)
    module = addon:NewModule(name,template,'AceHook-3.0',...)
    return module
end
